import json

JSONEncoder = json.JSONEncoder
JSONDecoder = json.JSONDecoder


class MyEncoder(JSONEncoder):
	"""MyDecoder hooks into :meth:`json.dump` and :meth:`json.dumps`"""

	def default(self, o):
		""""""
		if hasattr(o, '__json_encode__'):
			return o.__json_encode__()
		try:
			return JsonTypeRegister.encode(type(o).__name__, o)
		except KeyError:
			pass
		return super(MyEncoder, self).default(o)


class MyDecoder(JSONDecoder):
	"""MyDecoder hooks into :meth:`json.load` and :meth:`json.loads`"""

	def __init__(self, *args, **kwargs):
		super(MyDecoder, self).__init__(object_hook=self.object_hook, *args, **kwargs)

	def object_hook(self, dct):
		""""""
		if '__json_type__' in dct:
			return JsonTypeRegister.decode(dct['__json_type__'], dct)
		return dct


class JsonTypeRegister(object):
	"""Central location responsible for managing the mapping of class and class names to the appropriate encoders and
	decoders.

	:class:`JsonTypeRegister` has two sides:

	1. The user facing side  is for registering :ref:`encoders and decoders <encoderdecoder>`.

	2. The :mod:`json` facing side is to run a switch board for :class:`MyEncoder` and :class:`MyDecoder` connecting them to
	   the appropriate functions and methods for the type they are encoding / decoding in the moment.
	"""

	encoders = dict()
	decoders = dict()

	def __init__(self):
		super(JsonTypeRegister, self).__init__()

	@classmethod
	def register(cls, type_name: str, encoder: callable = None, decoder: callable = None):
		"""Helper for registering both an encode and custom_decoder in one call."""
		if encoder:
			cls.register_encoder(type_name, encoder)
		if decoder:
			cls.register_decoder(type_name, decoder)

	@classmethod
	def register_encoder(cls, type_name: str, custom_encoder: callable):
		"""Registers the custom_encoder against the type name.

		When the :mod:`json` encoder finds a type with name that has been registered, the custom_encoder will be called.
		The first argument will be the instance to be encoded.
		"""
		cls.encoders[type_name] = custom_encoder

	@classmethod
	def register_decoder(cls, type_name: str, custom_decoder: callable):
		"""Registers the custom_decoder against the type name.

		When the :mod:`json` decoder finds the type name in the serialized json, the custom_decoder will be called.
		The first argument will be the serialized json dict to be decoder.
		"""
		cls.decoders[type_name] = custom_decoder

	@classmethod
	def encode(cls, type_name, obj):
		"""Returns an encoded object if we have a registered handler or raise an exception."""
		if type_name in cls.encoders:
			return cls.encoders[type_name](obj)
		elif hasattr(obj, '__json_encode__'):
			return obj.__json_encode__()
		raise KeyError('Failed to find custom_encoder for {}'.format(obj))

	@classmethod
	def decode(cls, type_name, dct):
		"""Returns a decoded object if we have a registered handler or raise an exception."""
		if type_name in cls.decoders:
			return cls.decoders[type_name](dct)
		raise KeyError('Failed to find custom_decoder for {}'.format(dct))




"""
datetime
^^^^^^^^

:class:`datetime.datetime` is serialized as an isoformat string.

>>> import datetime
>>> data = datetime.datetime.utcnow()
>>> serialized = data.isoformat()
>>> serialized
'2021-04-07T00:00:08.114099'
>>> data = datetime.datetime.fromisoformat(serialized)
datetime.datetime(2021, 4, 7, 0, 0, 8, 114099)
"""

from ..cfjson import JsonTypeRegister


def datetime_decode(dct):
	"""Decode json dict into class objects."""
	from datetime import datetime
	cls_name = dct['__json_type__']
	if cls_name == 'datetime':
		return datetime.fromisoformat(dct['datetime'])
	raise TypeError()


def datetime_encode(obj):
	"""Encode the object into a json safe dict."""
	return {
		'__json_type__': 'datetime',
		'datetime': obj.isoformat()
	}


JsonTypeRegister.register('datetime', datetime_encode, datetime_decode)



"""
pathlib
^^^^^^^

:class:`pathlib.Path` is serialized as a string.

>>> import pathlib
>>> data = pathlib.Path('../source')
>>> serialized = str(data)
>>> serialized
'..\\source'
>>> data = pathlib.Path(serialized)
WindowsPath('../source')

Note: :mod:`pathlib` paths are platform specific. Special care must be taken if you are expecting the data to be
shared across systems.
"""

from ..cfjson import JsonTypeRegister


def path_decode(dct):
	"""Decode json dict into class objects."""
	from pathlib import Path
	cls_name = dct['__json_type__']
	if cls_name in ('Path', 'WindowsPath', 'PosixPath'):
		return Path(dct['path'])
	raise TypeError()


def path_encode(obj):
	"""Encode the object into a json safe dict."""
	return {
		'__json_type__': type(obj).__name__,
		'path': str(obj)
	}


JsonTypeRegister.register('Path', path_encode, path_decode)
JsonTypeRegister.register('WindowsPath', path_encode, path_decode)



"""
decimal
^^^^^^^

:class:`decimal.Decimal` is serialized as a string.

>>> import decimal
>>> data = decimal.Decimal('9.2')
>>> serialized = str(data)
>>> serialized
'9.2'
>>> data = decimal.Decimal(serialized)
Decimal('9.2')
"""

from ..cfjson import JsonTypeRegister


def decimal_decode(dct):
	"""Decode json dict into class objects."""
	from decimal import Decimal
	cls_name = dct['__json_type__']
	if cls_name == 'Decimal':
		return Decimal(dct['decimal'])
	raise TypeError()


def decimal_encode(obj):
	"""Encode the object into a json safe dict."""
	return {
		'__json_type__': 'Decimal',
		'decimal': str(obj)
	}


JsonTypeRegister.register('Decimal', decimal_encode, decimal_decode)


def decimal_test():

	from decimal import Decimal
	d = Decimal('1.23')
	from .. import dumps
	j = dumps({'blah': d})
	print(j)
	from .. import loads
	e = loads(j)
	print(e)


if __name__ == '__main__':
	decimal_test()



import json
from .cfjson import MyEncoder, MyDecoder, JsonTypeRegister

# Make sure all the serde get called and registered.
from .serde import *


JSONDecodeError = json.JSONDecodeError

__all__ = ('dump', 'dumps', 'load', 'loads', 'JSONDecodeError', 'JsonTypeRegister')
__version__ = '0.0.3.dev3'


def dump(obj, fp, *, skipkeys=False, ensure_ascii=True, check_circular=True,
        allow_nan=True, cls=None, indent=None, separators=None,
        default=None, sort_keys=False, **kw):
	"""Serialize ``obj`` as a JSON formatted stream to ``fp`` (a
	``.write()``-supporting file-like object).

	If ``skipkeys`` is true then ``dict`` keys that are not basic types
	(``str``, ``int``, ``float``, ``bool``, ``None``) will be skipped
	instead of raising a ``TypeError``.

	If ``ensure_ascii`` is false, then the strings written to ``fp`` can
	contain non-ASCII characters if they appear in strings contained in
	``obj``. Otherwise, all such characters are escaped in JSON strings.

	If ``check_circular`` is false, then the circular reference check
	for container types will be skipped and a circular reference will
	result in an ``OverflowError`` (or worse).

	If ``allow_nan`` is false, then it will be a ``ValueError`` to
	serialize out of range ``float`` values (``nan``, ``inf``, ``-inf``)
	in strict compliance of the JSON specification, instead of using the
	JavaScript equivalents (``NaN``, ``Infinity``, ``-Infinity``).

	If ``indent`` is a non-negative integer, then JSON array elements and
	object members will be pretty-printed with that indent level. An indent
	level of 0 will only insert newlines. ``None`` is the most compact
	representation.

	If specified, ``separators`` should be an ``(item_separator, key_separator)``
	tuple.  The default is ``(', ', ': ')`` if *indent* is ``None`` and
	``(',', ': ')`` otherwise.  To get the most compact JSON representation,
	you should specify ``(',', ':')`` to eliminate whitespace.

	``default(obj)`` is a function that should return a serializable version
	of obj or raise TypeError. The default simply raises TypeError.

	If *sort_keys* is true (default: ``False``), then the output of
	dictionaries will be sorted by key.
	"""
	kw['cls'] = MyEncoder
	return json.dump(obj, fp, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
        allow_nan=allow_nan, indent=indent, separators=separators,
        default=default, sort_keys=sort_keys, **kw)


def dumps(obj, *, skipkeys=False, ensure_ascii=True, check_circular=True,
        allow_nan=True, cls=None, indent=None, separators=None,
        default=None, sort_keys=False, **kw):
	"""Serialize ``obj`` to a JSON formatted ``str``.

	If ``skipkeys`` is true then ``dict`` keys that are not basic types
	(``str``, ``int``, ``float``, ``bool``, ``None``) will be skipped
	instead of raising a ``TypeError``.

	If ``ensure_ascii`` is false, then the return value can contain non-ASCII
	characters if they appear in strings contained in ``obj``. Otherwise, all
	such characters are escaped in JSON strings.

	If ``check_circular`` is false, then the circular reference check
	for container types will be skipped and a circular reference will
	result in an ``OverflowError`` (or worse).

	If ``allow_nan`` is false, then it will be a ``ValueError`` to
	serialize out of range ``float`` values (``nan``, ``inf``, ``-inf``) in
	strict compliance of the JSON specification, instead of using the
	JavaScript equivalents (``NaN``, ``Infinity``, ``-Infinity``).

	If ``indent`` is a non-negative integer, then JSON array elements and
	object members will be pretty-printed with that indent level. An indent
	level of 0 will only insert newlines. ``None`` is the most compact
	representation.

	If specified, ``separators`` should be an ``(item_separator, key_separator)``
	tuple.  The default is ``(', ', ': ')`` if *indent* is ``None`` and
	``(',', ': ')`` otherwise.  To get the most compact JSON representation,
	you should specify ``(',', ':')`` to eliminate whitespace.

	``default(obj)`` is a function that should return a serializable version
	of obj or raise TypeError. The default simply raises TypeError.

	If *sort_keys* is true (default: ``False``), then the output of
	dictionaries will be sorted by key.

	To use a custom ``JSONEncoder`` subclass (e.g. one that overrides the
	``.default()`` method to serialize additional types), specify it with
	the ``cls`` kwarg; otherwise ``JSONEncoder`` is used.

	"""
	kw['cls'] = MyEncoder
	return json.dumps(obj, skipkeys=skipkeys, ensure_ascii=ensure_ascii, check_circular=check_circular,
        allow_nan=allow_nan, indent=indent, separators=separators,
        default=default, sort_keys=sort_keys, **kw)


def load(fp, *, cls=None, object_hook=None, parse_float=None,
        parse_int=None, parse_constant=None, object_pairs_hook=None, **kw):
	"""Deserialize ``fp`` (a ``.read()``-supporting file-like object containing
	a JSON document) to a Python object.

	``object_hook`` is an optional function that will be called with the
	result of any object literal decode (a ``dict``). The return value of
	``object_hook`` will be used instead of the ``dict``. This feature
	can be used to implement custom decoders (e.g. JSON-RPC class hinting).

	``object_pairs_hook`` is an optional function that will be called with the
	result of any object literal decoded with an ordered list of pairs.  The
	return value of ``object_pairs_hook`` will be used instead of the ``dict``.
	This feature can be used to implement custom decoders.  If ``object_hook``
	is also defined, the ``object_pairs_hook`` takes priority.
	"""
	kw['cls'] = MyDecoder
	return json.load(fp, object_hook=object_hook, parse_float=parse_float,
        parse_int=parse_int, parse_constant=parse_constant, object_pairs_hook=object_pairs_hook **kw)


def loads(s, *, cls=None, object_hook=None, parse_float=None,
        parse_int=None, parse_constant=None, object_pairs_hook=None, **kw):
	"""Deserialize ``s`` (a ``str``, ``bytes`` or ``bytearray`` instance
	containing a JSON document) to a Python object.

	``object_hook`` is an optional function that will be called with the
	result of any object literal decode (a ``dict``). The return value of
	``object_hook`` will be used instead of the ``dict``. This feature
	can be used to implement custom decoders (e.g. JSON-RPC class hinting).

	``object_pairs_hook`` is an optional function that will be called with the
	result of any object literal decoded with an ordered list of pairs.  The
	return value of ``object_pairs_hook`` will be used instead of the ``dict``.
	This feature can be used to implement custom decoders.  If ``object_hook``
	is also defined, the ``object_pairs_hook`` takes priority.

	``parse_float``, if specified, will be called with the string
	of every JSON float to be decoded. By default this is equivalent to
	float(num_str). This can be used to use another datatype or parser
	for JSON floats (e.g. decimal.Decimal).

	``parse_int``, if specified, will be called with the string
	of every JSON int to be decoded. By default this is equivalent to
	int(num_str). This can be used to use another datatype or parser
	for JSON integers (e.g. float).

	``parse_constant``, if specified, will be called with one of the
	following strings: -Infinity, Infinity, NaN.
	This can be used to raise an exception if invalid JSON numbers
	are encountered.

	The ``encoding`` argument is ignored and deprecated since Python 3.1.
	"""
	kw['cls'] = MyDecoder
	return json.loads(s, object_hook=object_hook, parse_float=parse_float,
        parse_int=parse_int, parse_constant=parse_constant, object_pairs_hook=object_pairs_hook, **kw)
